#
# The MIT License (MIT)
# Copyright (c) 2022 M.O.S.S. Computer Grafik Systeme GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the Software
# is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT,TORT OR OTHERWISE, ARISING FROM, OUT
# OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import json
import logging
import os
from pathlib import Path
from typing import Optional

from moss.ems.emsproject import EmsProject
from moss.ems.emsservice import EmsServiceException, Service
from moss.ems.extra.geometry_converter import GeojsonConverter

logger = logging.getLogger("ems_cli_export")

try:
    from osgeo import gdal
except ImportError:
    logger.error("This function need GDAL Python bindings.")
    raise ImportError("No GDAL Provided")


class EmsCliLoadException(Exception):
    pass


class EmsCliLoad:
    def __init__(
        self,
        url: str,
        project: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
    ):

        if username and password:
            logger.debug("Logging in using username and password.")
            try:
                self.service = Service(url, username=username, password=password)
            except EmsServiceException:
                logger.error("Can not connect to service using username and password.")
                raise EmsCliLoadException
        else:
            logger.debug("Using service without authentication.")
            try:
                self.service = Service(url)
            except EmsServiceException:
                logger.error("Can not access WEGA-EMS Service using no authentication.")
        logger.debug("Accessing project %s", project)
        try:
            selected_project = self.service.project(project)
            if selected_project is not None:
                self.project: EmsProject = selected_project
            else:
                logger.error("The selected project %s is not defined", project)
                raise EmsCliLoadException
        except EmsServiceException:
            logger.error("The project %s does not exists in %s", project, url)
            raise EmsCliLoadException

    def ems_load(
        self,
        input_path: str,
        object_class: str,
        layer_name: str,
        variant_id: Optional[int] = 11,
    ):

        logger.info("Starting load...")
        self.objectclass = object_class
        self.layername = layer_name
        my_project = self.project
        logger.info("ObjectClass: {0}".format(object_class))
        logger.info("Input Path: {0}".format(input_path))
        logger.info("Layer Name: {0}".format(layer_name))
        logger.info("Variant ID: {0}".format(variant_id))

        if not Path(input_path).exists():
            raise EmsCliLoadException("The output path %s not exists.", input_path)

        for file in os.listdir(input_path):
            if file.endswith(".gpkg") or file.endswith(".shp"):
                shape_path = Path(input_path) / file
                file_name = file.split(".")[0]
                json_name = "{0}.{1}".format(file_name, "json")
                geojson_path = Path(input_path) / str(json_name)

                # shape to geojson
                gdal.VectorTranslate(
                    str(geojson_path),
                    str(shape_path),
                    layerName=file_name,
                    format="GeoJSON",
                )

                # geojson to esri
                with open(geojson_path, "r") as geo_js:
                    data_geojson = json.loads(geo_js.read())

                # print (data_geojson)
                esri_output = GeojsonConverter.to_esri(data_geojson["features"][0])

                esri_name = "{0}_{1}.{2}".format(file_name, "esri", "json")

                esri_path = Path(input_path) / str(esri_name)

                with open(esri_path, "w") as esri_f:
                    json.dump(esri_output, esri_f)

                my_layer = my_project.objectclass(self.objectclass).layer(
                    self.layername
                )

                # sending to ems
                output = my_layer.add_features(
                    features=[esri_output],
                    variant_id=variant_id,
                )
                # response
                logger.info("Respond from ems: %s", output)
                geojson_path.unlink()
                esri_path.unlink()

        logger.info("Closing the communication with WEGA-EMS")
        self.service.close()



#
# The MIT License (MIT)
# Copyright (c) 2022 M.O.S.S. Computer Grafik Systeme GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the Software
# is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT,TORT OR OTHERWISE, ARISING FROM, OUT
# OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import json
import logging

from ..emscatalog import EmsCatalog

logger = logging.getLogger(__file__)


class EmsPaginationQueryException(Exception):
    """
    Exception for EmsLayer
    """


class PaginationQuery:
    """
    This will run a query using pagination.
    """

    def __init__(self, session, url, project, **kwargs):
        self.session = session
        self.url = url
        self.project = project
        self.query_parameters = kwargs

        # for key, value in kwargs.items():
        #    setattr(self, key, value)

        self.length = self._result_length()
        if self.length < 0:
            raise EmsPaginationQueryException("An error occured in querying the layer.")

        self._saved_catalogs = {}

    def _result_length(self):
        """
        A query to get the total number of results
        """
        params = {
            "f": "json",
            "returnCountOnly": True,
            "where": self.query_parameters["where"],
        }

        if "variants" in self.query_parameters:
            params["variants"] = self.query_parameters["variants"]

        if "geometry" in self.query_parameters:
            params["geometry"] = json.dumps(self.query_parameters["geometry"])
            params["geometryType"] = "esriGeometryEnvelope"

        response = self.session.get(self.url, params=params)
        response_dict = response.json()

        if "count" not in response_dict:
            error_dictionary = response_dict["error"]
            logger.error("Message: %s", error_dictionary["message"])
            logger.error("Description: %s", error_dictionary["description"])
            logger.error("Code: %s", error_dictionary["code"])

        return response_dict.get("count", -1)

    def __len__(self):
        if isinstance(self.length, str):
            return 0
        else:
            return self.length

    def _catalog(self, catalog_name):
        """
        Return the catalog.
        """

        url = "/".join(["rest", self.project, catalog_name, "query"])
        request_parameters = {"f": "json", "query": "1=1"}
        response = self.session.get(url, params=request_parameters)
        response_dict = response.json()
        # logger.debug("Response catalog %s", response_dict)
        catalog = EmsCatalog(
            name=catalog_name, session=response_dict, entries=response_dict["entries"]
        )

        return catalog

    def _process_catalog(self, query_results):
        """
        Replace the catalogs values in the result
        """

        # Using the fields we create a dictionary with as key the field name
        # and as value the catalog where this entry exist
        field_catalog = {
            field["name"]: field["catalogName"]
            for field in query_results["fields"]
            if "catalogName" in field
        }

        for key, feature in enumerate(query_results["features"]):
            for attribute, value in list(feature["attributes"].items()):

                logger.debug(
                    "Current attribute is %s current value is %s", attribute, value
                )
                if attribute in field_catalog:

                    catalog_name = field_catalog[attribute]
                    logger.debug("Catalog name %s", catalog_name)

                    if catalog_name in self._saved_catalogs.keys():
                        logger.debug("Catalog already stored")
                        catalog = self._saved_catalogs[catalog_name]
                    else:
                        catalog = self._catalog(catalog_name)
                        self._saved_catalogs[catalog_name] = catalog

                        logger.debug("Stored catalog %s", self._saved_catalogs)

                    # Update the values with the values in catalog
                    query_results["features"][key]["attributes"][attribute] = catalog[
                        value
                    ]
        return query_results

    def resolve(self, page_size=500, with_catalog=False):
        """
        Run the query to get the result

        :param page_size: The number of records to fetch.
        :type page_size: int
        :param with_catalog: A switch to replace the catalog values of the result
        :type with_catalog: bool
        :returns:  the data of the pagination query according to the parameters set
        :raises:
        """
        record_offset = 0
        if "geometry" in self.query_parameters:
            self.query_parameters["geometry"] = json.dumps(
                self.query_parameters["geometry"]
            )
            self.query_parameters["geometryType"] = "esriGeometryEnvelope"

        while True:
            pagination_params = {
                "resultRecordCount": page_size,
                "resultOffset": record_offset,
            }

            params = self.query_parameters.copy()
            params.update(pagination_params)

            response = self.session.post(self.url, data=params)

            response.raise_for_status()

            data = response.json()

            if record_offset > self.length:
                break

            if with_catalog:

                yield self._process_catalog(data)
            else:
                yield data

            record_offset += page_size



#
# The MIT License (MIT)
# Copyright (c) 2022 M.O.S.S. Computer Grafik Systeme GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the Software
# is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT,TORT OR OTHERWISE, ARISING FROM, OUT
# OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import base64
import json
import logging

import requests

from .emsproject import EmsProject
from .utilities import SessionBaseUrl

logger = logging.getLogger("ems_service")


__version__ = "2.0.0rc1"


class EmsServiceException(Exception):
    """
    Handles general exceptions for the service
    """


class EmsServiceAuthException(Exception):
    """
    Handles error related to authentication. For example
    an expired token.
    """


class Service(object):
    """
    A Class used to represent an EMS Service.

    To access a WEGA-EMS Service without authentication:

    Example:
        >>> my_service = Service("http://localhost:8080/wega-ems/")
        >>> my_service.projects
        ["project1", "project2"]

    If the service uses authentication, username and password can be provided
    as additional parameters:

    Example:
        >>> my_service = Service("http://localhost:8080/wega-ems/",
                username="Test",
                password="Test")
        >>> my_service.projects
        ["project1", "project2"]

    If working wih token, you can easily ignore the username and password:

    Example:
        >>> my_service = Service("http://localhost:8080/wega-ems/",
                                token="aabbcc")
        >>> my_service.projects
        ["project1", "project2"]

    Emspy uses requests Session so the session will remain active until close()
    is called.This will perform also a logout on Tomcat.

    Args:

        url (str):      The url to the WEGA-EMS instance service.
        username (str, optional): The username used for authentication.
        password (str, optional): The password used for authentication.
        token (str, optional): The token used for authentication
    """

    def __init__(self, base_url, username=None, password=None, token=None):
        # type: (str, str, str, str) -> None
        if not base_url.endswith("/"):
            base_url = base_url + "/"
        self.base_url = base_url
        self.username = username
        self.password = password

        if not self._url_is_correct():
            logger.error("The provided url is probably wrong.")
            raise EmsServiceException("The provided url is probably wrong.")

        # Create the session used to run the API call
        self.session = SessionBaseUrl(base_url)

        # Check if the service is secured. If so we need a username
        # and a password.
        if self._service_is_secured():
            logger.debug("The service is secured.")
            if token is not None:
                logger.debug("Using token %s", token)
                bearer_token = base64.urlsafe_b64encode(token.encode()).decode("utf-8")
                bearer_string = "Bearer {}".format(bearer_token)
                self.session.headers.update({"Authorization": bearer_string})
            else:
                if self.username:
                    if not self._login_user():
                        raise EmsServiceException("Wrong credentials provided.")

                else:
                    logger.error("The service is secured and there is no token.")
                    raise EmsServiceException(
                        "The service is secured. You need to provide a token."
                    )

        self._parameters = self._get_parameters()

    def _login_user(self):
        """
        This is the internal function to log in.
        """

        url = "login.xhtml"

        logger.info("Logging user using the address %s", url)
        payload = {"username": self.username, "password": self.password}
        self.session.post(url, data=payload)

        # Test if we can read the projects. Just to see if the log is ok.
        # If the password is wrong, the code is always 200.

        url = "/".join(["rest", "info"])
        parameters = {"f": "json"}

        response = self.session.get(url, params=parameters)

        if response.status_code == 200:
            try:
                response_dict = json.loads(response.text)
            except ValueError as exception:
                logging.error(
                    "Login error. Please check username and password: %s", exception
                )
                return False
            else:
                self.information = response_dict
                return True
        else:
            raise EmsServiceException("Error logging in")

    def _get_parameters(self):
        """
        This will get the parameters of the current Project
        """
        url = "/".join(["rest", "projects"])
        parameters = {"f": "json"}

        response_dict = self._ems_request(url, parameters)

        return response_dict

    def _url_is_correct(self):
        """
        Check if the url is correct
        """
        url = "/".join([self.base_url, "rest"])
        logger.info("Url %s", url)

        if "jsessionid" in url:
            url = url.split(";")[0]
        try:
            response = requests.get(url)
        except (ValueError, requests.exceptions.ConnectionError) as exception:
            logging.error(
                "Error connecting to EMS Service for URL Testing: %s", exception
            )
            return False
        else:
            return response.status_code == 200 or response.status_code == 400

    def _service_is_secured(self):
        """
        This one wil return true if the service can be used
        without token or username/password
        """
        url = "/".join(["rest", "projects"])
        parameters = {"f": "json"}

        logging.debug("Checking if service at %s is secured", url)

        response = self.session.get(url, params=parameters)
        logging.debug(" Status code: %s", response.status_code)

        if "application/json" in response.headers["Content-Type"]:
            logging.info("The service needs no authentication.")
            return False

        return True

    @property
    def info(self):
        """
        List of all projects with some meta data.

        Examples:

            >>> my_service = Service("http://localhost:8080/wega-ems/")
            >>> my_service.info

        Returns:
            list: A list of projects and meta data for the current service
        """
        url = "/".join(["rest", "info"])
        parameters = {"f": "json"}

        information = self._ems_request(url, parameters)

        return information

    @property
    def configuration(self):
        """
        Return the base configuration of an administration client.

        Returns:
            a configuration dictionary containing a
            success property and version
        """
        url = "/".join(["rest", "configuration"])
        request_parameters = {"f": "json"}

        response_dict = self._ems_request(url, request_parameters)

        logger.info("Configuration loaded successfully for the service")
        return response_dict

    @property
    def projects(self):
        """
        The list of all the projects in the current Service.

        Example:

            >>> my_service = Service("http://localhost:8080/wega-ems/")
            >>> my_service.projects
            ['project1', 'project2']

        Returns:
           list: A list of `EmsProject`
        """
        url = "/".join(["rest", "projects"])
        parameters = {"f": "json"}

        response = self._ems_request(url, parameters)
        try:
            projects = response["projects"]
        except KeyError:  # pragma: no cover
            logger.debug("Getting projects %s", response)
            raise EmsServiceException("There is no projects key in response")
        return [EmsProject(self.session, project) for project in projects]

    def add_project(self, project_name, description=""):
        """
        Add a new Project to the current Service.

        Args:
            project_name (str): The project name
            description (str, optional): The project description.
            If not provided it will be an empty string.

        Example:
            >>> my_service = Service("http://localhost:8080/wega-ems/")
            >>> my_service.add_project("PROJECT", "Beschreibung")

        Returns
            The `EmsProject` instance of the added project
            or None in case of error
        """
        logger.debug("Trying to add the project '" + project_name + "'.")

        url = "/".join(["rest", "addProject"])
        request_parameters = {
            "f": "json",
            "name": project_name,
            "description": description,
        }

        response_dict = self._ems_request(url, request_parameters, "POST")

        try:
            if response_dict["addProjectResult"]["success"] is False:
                error_dictionary = response_dict["addProjectResult"]["error"]

                logger.error("Message: %s", error_dictionary["message"])
                logger.error("Description: %s", error_dictionary["description"])
                logger.error("Code: %s", error_dictionary["code"])
                return None
        except KeyError:  # pragma: no cover
            logger.error(
                "add_project delivered an unexpected response: %s", response_dict
            )
            raise EmsServiceException(
                "add_project delivered an unexpected response: %s", response_dict
            )

        logger.info("Added successfully the project '" + project_name + "'.")
        project = self.project(project_name)
        return project

    def delete_project(self, project_name):
        """
        Deletes the provided project. This operation can not be undone.

        Args:
            project_name (str): Project name

        Examples:

            >>> my_service = Service("http://localhost:8080/wega-ems/")
            >>> my_service.delete_project("PROJECT")

        Returns:
            bool: `True` or `False` if the project deleted successfully.

        """
        projects = self.projects

        # Choose the project
        project = next(iter(filter(lambda p: p.name == project_name, projects)), None)

        if project is None:
            logging.error(
                "The project %s does not exist in this instance", project_name
            )
            raise EmsServiceException(
                "The project %s does not exist in this instance" % project_name
            )

        project_id = project.id  # type: ignore

        # Send request
        url = "/".join(["rest", "deleteProject"])
        request_parameters = {"f": "json", "projectId": project_id}

        response_dict = self._ems_request(url, request_parameters, "POST")

        try:
            if response_dict["deleteProjectResult"]["success"] is False:
                logger.error(
                    "Message: %s",
                    response_dict["deleteProjectResult"]["error"]["message"],
                )
                logger.error(
                    "Description: %s",
                    response_dict["deleteProjectResult"]["error"]["description"],
                )
                logger.error(
                    "Code: %s", response_dict["deleteProjectResult"]["error"]["code"]
                )
                return False
        except KeyError:  # pragma: no cover
            logger.error(
                "delete_project delivered an unexpected response: %s", response_dict
            )
            raise EmsServiceException(
                "delete_project delivered an unexpected response: %s", response_dict
            )

        logger.info("Deleted successfully the project")
        return True

    def project(self, project_name):
        """
        Returns a project class by the project name.

        Example:
            >>> my_service.project("prj1")

        Args:
            project_name: The name of the project

        Returns:
            :obj:`EmsProject`
            or None if the project could not be found
        """

        projects = self.projects

        project = next(iter(filter(lambda p: p.name == project_name, projects)), None)

        if project is not None:
            return project

        else:
            logging.error(
                "The project %s does not exist in this instance", project_name
            )

    def export_project(self, project_name):
        """
        Exports the provided project

        Example:
            >>> my_service.export_project("prj1")

        Args:
            project_name: The name of the project to be exported

        Returns:
            :dict: Exported project information
        """
        logger.info("Starting the export of the project structure.")

        url = "/".join(["rest", project_name, "saveProject"])
        logger.debug("Export project %s", url)

        request_parameters = {
            "f": "json",
        }
        response_dict = self._ems_request(url, request_parameters, "POST")

        logger.debug("Project structure exported %s", response_dict)

        return response_dict

    def import_project(
        self,
        project_definition,
        project_name,
        password=None,
    ):
        """
        Import a project using the provided dict

        Example:
            >>> my_service.import_project("prj1")

        Args:
            project_definition (dict): The name of the project to be imported
            project_name       (str): The name of the project
            password           (str): The password

        Returns:
            bool: The return value. True for success, False otherwise.
        """
        url = "/".join(["rest", "loadProject"])

        files = {
            "projectFile": (
                "loaded_project.json",
                json.dumps(project_definition),
                "application/json",
            ),
            "name": (None, project_name),
        }

        if password is not None:
            files["password"] = (None, password)

        response = self.session.post(url, files=files)
        response_dict = response.json()
        logger.debug("Response from import %s", response_dict)

        result = response_dict.get("loadProjectResult")

        if result is not None:
            if result["success"] is True:
                logger.info("Imported the project structure successfully.")
                return True
            else:
                logger.error("An error occured while importing the project structure!")
                logger.error("Message: %s", result["error"]["message"])
                logger.error("Description: %s", result["error"]["description"])
                logger.error("Code: %s", result["error"]["code"])
                return False
        else:
            logger.error(
                "Error loading project: %s", response_dict["loadProjectResult"]["error"]
            )
            return False

    def _ems_request(self, url, parameters, method="GET"):
        """Return a json response from a session"""
        logger.debug("New request to %s parameters %s", url, parameters)

        if method == "POST":
            response = self.session.post(url, data=parameters, allow_redirects=False)
        else:
            response = self.session.get(url, params=parameters, allow_redirects=False)

        # Handle redirect to login page

        if response.status_code == 302:
            raise EmsServiceAuthException("Invalid token provided")

        if response.status_code != 200:  # pragma: no cover
            logger.error(
                "Response is %s. Url: %s Parameters:%s Status code: %s",
                response.status_code,
                url,
                parameters,
                response.status_code,
            )
            raise EmsServiceException(
                "Communication error with EMS"
            )  # pragma: no cover
        try:
            json_response = response.json()
        except ValueError as exception:  # pragma: no cover
            logger.exception(
                "Response is not json. You are probablyon the login website:%s",
                exception,
            )
        else:
            return json_response

    def close(self):
        """
        Close the session to the service.
        """
        logout_url = "/".join([self.base_url, "logout.xhtml"])
        response = self.session.get(logout_url)
        if response.ok:
            logger.info("Session closed.")
        else:
            logger.warn("Error logging out using url %s", logout_url)
        self.session.close()
        logger.info("Session closed.")



#
# The MIT License (MIT)
# Copyright (c) 2022 M.O.S.S. Computer Grafik Systeme GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the Software
# is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT,TORT OR OTHERWISE, ARISING FROM, OUT
# OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import copy
import json
import logging
from pathlib import Path
from typing import List, Optional

from moss.ems.emslayer import EmsLayer
from moss.ems.emsproject import EmsProject
from moss.ems.emsservice import EmsServiceException, Service

logger = logging.getLogger("ems_cli_export")

try:
    from osgeo import gdal
except ImportError:
    logger.error("This function needs GDAL Python bindings.")
    raise ImportError("No GDAL Provided")


class EmsCliDumpException(Exception):
    pass


class EmsCliDump:
    """ """

    def __init__(
        self,
        url: str,
        project: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        token: Optional[str] = None,
    ):

        if username and password:
            logger.debug("Logging in using username and password.")
            try:
                self.service = Service(url, username=username, password=password)
            except EmsServiceException:
                logger.error("Can not connect to service using username and password.")
                raise EmsCliDumpException
        elif token:
            logger.debug("Logging in using token.")
            try:
                self.service = Service(url, token=token)
            except EmsServiceException:
                logger.error("Can not connect to service using token.")
                raise EmsCliDumpException
        else:
            logger.debug("Using service without authentication.")
            try:
                self.service = Service(url)
            except EmsServiceException:
                logger.error("Can not access WEGA-EMS Service using no authentication.")

        logger.debug("Accessing project %s", project)
        try:
            selected_project = self.service.project(project)
            if selected_project is not None:
                self.project: EmsProject = selected_project
            else:
                logger.error("The selected project %s is not defined", project)
                raise EmsCliDumpException
        except EmsServiceException:
            logger.error("The project %s does not exists in %s", project, url)
            raise EmsCliDumpException

    def _project_has_variant(self):
        """
        Check if project has variants
        """
        return self.project.variants_tree()  # Empty dict are False

    def ems_dump(
        self,
        output_path: str,
        variant_master_name: Optional[str] = None,
        variant_master_id: Optional[str] = None,
        objectclasses: Optional[str] = None,
        variants: Optional[str] = None,
        extension: str = "gpkg",
        output_filename: str = "variant",
    ):
        logger.info("Starting dump..")
        if variant_master_id is not None or variant_master_name is not None:
            logger.info("Starting dump using variants.")
            logger.info(
                "Variant master name: %s - Variant master id: %s ",
                variant_master_name,
                variant_master_id,
            )

        if variant_master_id and variant_master_name:
            logger.error("Please provide master name or the id, not booth.")
            raise EmsCliDumpException

        if not Path(output_path).exists():
            raise EmsCliDumpException("The output path %s not exists.", output_path)

        variants_tree, _ = self.project.variants_tree()

        selected_variant = {}

        try:
            variants_list = variants_tree["children"]
        except KeyError:
            logger.error("Error parsing variant Tree")
            raise EmsCliDumpException

        if variant_master_name:
            selected_variant = next(
                variant
                for variant in variants_list
                if variant["variant_name"] == variant_master_name
            )
            logger.debug(
                "Found this variant using name %s as selection: %s",
                variant_master_name,
                selected_variant,
            )

        if variant_master_id:
            selected_variant = next(
                variant
                for variant in variants_list
                if variant["id"] == int(variant_master_id)
            )
            logger.debug(
                "Found this variant using id %s as selection: %s",
                variant_master_id,
                selected_variant,
            )

        if selected_variant is None:
            logger.error("The selected variant can not be found.")
            raise EmsCliDumpException

        logger.info(
            "Starting export all the sub-variants inside %s",
            output_path,
        )

        exported_variants = []
        if variants is not None:
            exported_variants = variants.split(",")
        else:
            exported_variants = selected_variant["children"]

        for variant in exported_variants:

            try:
                variant_id = variant["id"]
            except TypeError:
                variant_id = int(variant)
                if variant_id == int(variant_master_id):
                    continue

            logger.info("Current variant id %s", variant_id)

            variant_geopackage = "{0}-{1}.{2}".format(
                output_filename, variant_id, extension
            )
            variant_geopackage_path = Path(output_path) / variant_geopackage

            export_objectclasses = []
            if objectclasses is not None:
                export_objectclasses = objectclasses.split(",")
            else:
                export_objectclasses = self.project.objectclasses

            for objectclass in export_objectclasses:
                logger.info("Processing objectclass %s", objectclass)

                if isinstance(objectclass, str):
                    objectclass = self.project.objectclass(objectclass)

                layers: Optional[List[EmsLayer]] = objectclass.layers
                if layers is not None:

                    for layer in layers:
                        logger.debug("Processing layer %s", layer)

                        layer_query = layer.query(
                            where="1=1", returnGeometry=True, variants=[variant_id]
                        )
                        if layer_query:
                            query_output = layer_query.resolve(with_catalog=True)

                            total_features = []
                            final_esri = {}

                            for query_index, query_item in enumerate(query_output):
                                if query_index == 0:
                                    final_esri = copy.deepcopy(query_item)

                                features = query_item.get("features")
                                if features is not None:
                                    total_features.extend(features)

                            cleaned_features = list(
                                filter(lambda item: "geometry" in item, total_features)
                            )

                            if cleaned_features:
                                final_esri["features"] = cleaned_features
                                final_esri["counts"] = len(cleaned_features)

                                temp_layer = "temp-{0}-{1}.json".format(
                                    objectclass, layer
                                )
                                temp_layer_path = Path(output_path) / temp_layer
                                temp_layer_name = temp_layer.split(".")[0]

                                with open(temp_layer_path, "w+") as esri_json:
                                    esri_json.write(json.dumps(final_esri))

                                if extension == "csv":
                                    csvpackage = "{0}.{1}".format(str(layer), extension)
                                    variant_dirname = "{0}-{1}".format(
                                        output_filename, variant_id
                                    )
                                    variant_path = Path(output_path) / Path(
                                        variant_dirname
                                    )
                                    variant_path.mkdir(parents=True, exist_ok=True)
                                    variant_geopackage_path = variant_path / csvpackage

                                logger.info(
                                    "Exporting current variant in %s",
                                    variant_geopackage_path,
                                )

                                if not variant_geopackage_path.exists():
                                    gdal.VectorTranslate(
                                        str(variant_geopackage_path),
                                        str(temp_layer_path),
                                        layerName=temp_layer_name,
                                    )
                                else:
                                    gdal.VectorTranslate(
                                        str(variant_geopackage_path),
                                        str(temp_layer_path),
                                        layerName=temp_layer_name,
                                        accessMode="update",
                                    )

                                temp_layer_path.unlink()

        logger.info("Closing the communication with WEGA-EMS")
        self.service.close()



import logging

logger = logging.getLogger("ems_cli_compare")


class EmsCliCompare:
    def ems_compare(self, project_name1, project_name2):
        # type: (str, str) -> None
        """Compares two EMS projects

        Args:
            project_name1: the name of the first EMS project
            project_name2: the name of the second EMS project

        Returns:

        Raises:
            EmsCliException
        """
        from moss.ems.scripts.ems_cli import EmsCliException

        logger.info(
            "Start comparing project '{}' with project '{}'".format(
                project_name1, project_name2
            )
        )
        logger.debug(
            "Projects in the current instance: %s",
            self.ems_service.projects,  # type: ignore
        )
        error_counter = 0
        error_log = []

        project1 = next(
            iter(filter(lambda p: p.name == project_name1, self.ems_service.projects)),  # type: ignore
            None,
        )

        if project1 is None:
            logger.error(
                "The first project '%s' does not exist in this instance", project_name1
            )
            raise EmsCliException(
                "The first project '{}' does not exist in this instance".format(
                    project_name1
                )
            )

        project2 = next(
            iter(filter(lambda p: p.name == project_name2, self.ems_service.projects)),  # type: ignore
            None,
        )

        if project2 is None:
            logger.error(
                "The second project '%s' does not exist in this instance", project_name2
            )
            raise EmsCliException(
                "The second project '{}' does not exist in this instance".format(
                    project_name2
                )
            )

        # Check if the project has variant
        objectclasses1 = project1.objectclasses
        # Looking for the VNT Master Objectclass
        variant_master_found1 = None
        variant_master_found1 = next(
            (item for item in objectclasses1 if item.objectClassType == "VNTMASTER"), None  # type: ignore
        )
        # Check if the project has variant
        objectclasses2 = project2.objectclasses
        # Looking for the VNT Master Objectclass
        variant_master_found2 = None
        variant_master_found2 = next(
            (item for item in objectclasses2 if item.objectClassType == "VNTMASTER"), None  # type: ignore
        )
        logger.info("Starting Compare")
        if variant_master_found1:
            logger.info(
                "An objectclass of type 'VNTMASTER' was found in {}. Switching to"
                " variant mode!".format(project1)
            )
            if variant_master_found2:
                logger.info(
                    "An objectclass of type 'VNTMASTER' was also found in {}.".format(
                        project2
                    )
                )

            query_results1 = variant_master_found1.layers[0].query()
            query_results2 = variant_master_found2.layers[0].query()

            featureid_name1 = list(query_results1.resolve())[0]["objectIdFieldName"]
            featureid_name2 = list(query_results2.resolve())[0]["objectIdFieldName"]
            for query1 in list(query_results1.resolve())[0]["features"]:
                # since an error count greater than 100 continuation makes no sense
                if error_counter >= 100:
                    break
                featureid1 = query1["attributes"][featureid_name1]
                if (
                    len(
                        [
                            q
                            for q in list(query_results2.resolve())[0]["features"]
                            if q["attributes"]["GUID"] == query1["attributes"]["GUID"]
                        ]
                    )
                    > 1
                ):
                    logger.error(
                        "GUID {} is not unique".format(query1["attributes"]["GUID"])
                    )
                query2 = next(
                    (
                        q
                        for q in list(query_results2.resolve())[0]["features"]
                        if q["attributes"]["GUID"] == query1["attributes"]["GUID"]
                    ),
                    None,
                )
                featureid2 = query2["attributes"][featureid_name2]
                logger.info(
                    "Comparing master feature {}, comparing with {}".format(
                        featureid1, featureid2
                    )
                )
                variants1 = variant_master_found1.variants(featureid1)
                variants2 = variant_master_found2.variants(featureid2)
                for variant1 in variants1:
                    if error_counter >= 100:
                        break
                    _id1 = variant1[featureid_name1]
                    if len([v for v in variants2 if v["NAME"] == variant1["NAME"]]) > 1:
                        logger.error(
                            "Variant name {} is not unique".format(variant1["NAME"])
                        )
                    variant2 = next(
                        (v for v in variants2 if v["NAME"] == variant1["NAME"]), None
                    )
                    _id2 = variant2[featureid_name2]
                    logger.info(
                        "Comparing variant {} comparing with {}".format(_id1, _id2)
                    )
                    for objectclass in objectclasses1:
                        if error_counter >= 100:
                            break
                        if objectclass.has_variant:
                            logger.info(
                                "Comparing objectclass {}".format(objectclass.name)
                            )
                            for layer in objectclass.layers:
                                logger.info(
                                    "Comparing layer data for layer {}".format(
                                        layer.name
                                    )
                                )
                                layer_data1 = layer.query(variants=[_id1])
                                layer_data2 = (
                                    project2.objectclass(objectclass.name)
                                    .layer(layer.name)
                                    .query(variants=[_id2])
                                )
                                # features1 = list(layer_data1.resolve())
                                # features2 = list(layer_data2.resolve())
                                if layer_data1.length != layer_data2.length:
                                    error_message = (
                                        "Feature list count of {}/{}/{}/{} is NOT equal"
                                        " ({} != {})!".format(
                                            project1,
                                            variant1["NAME"],
                                            objectclass.name,
                                            layer.name,
                                            layer_data1.length,
                                            layer_data2.length,
                                        )
                                    )
                                    logger.error(error_message)
                                    error_log.append(error_message)
                                    error_counter += 1
                                    if error_counter >= 100:
                                        logger.error(
                                            "Too many differences: stopping the"
                                            " compare."
                                        )
                                        break

        else:
            logger.info("No VARIANTMASTER found.")

            for objectclass in objectclasses1:
                logger.debug("Comparing objectclass %s", objectclass)
                for layer in objectclass.layers:
                    logger.debug("Comparing layer data for layer %s", layer)
                    layer_data1 = layer.query()
                    layer_data2 = (
                        project2.objectclass(objectclass.name).layer(layer.name).query()
                    )
                    if layer_data1.length != layer_data2.length:
                        error_message = (
                            "Feature list count of {}/{}/{} is NOT equal ({} != {})!"
                            .format(
                                project1,
                                objectclass.name,
                                layer.name,
                                layer_data1.length,
                                layer_data2.length,
                            )
                        )
                        logger.error(error_message)
                        error_log.append(error_message)
                        error_counter += 1
                        if error_counter >= 100:
                            logger.error("Too many differences: stopping the compare.")
                            break
        if error_counter > 0:
            logger.info(
                "Compare completed with {} errors!\nError log summary:\n".format(
                    error_counter
                )
            )
            logger.error("\n".join(map(str, error_log)))
        else:
            logger.info("Compare completed successfully.")



#
# Copyright (c) 2020  by M.O.S.S. Computer Grafik Systeme GmbH
#                         Hohenbrunner Weg 13
#                         D-82024 Taufkirchen
#                         http://www.moss.de#


class ObjectClassSorter:
    """
    Hierarchical ObjectClassSorter

    Examples:
        Builds an objectclass sorter

        >>> sorter = ObjectClassSorter(objectclasses)
        >>> result = sorter.by_reference()
    """

    def __init__(self, objectclasses):
        """
        Creates a new ObjectClassSorter instance.

        Args:
            objectclasses: The list of ObjectClasses to sort
        """
        self.objectclasses = objectclasses

    def by_reference(self):
        """
        Sorts the ObjectClasses by reference.
        """

        collected = []

        for root in self._get_root():
            collected.extend(self._collect(root))

        return collected

    def _collect(self, objectclass):
        collected = []

        collected.append(objectclass)

        for child in self._get_children(objectclass):
            collected.extend(self._collect(child))

        return collected

    def _get_children(self, objectclass):
        objectclasses = filter(
            lambda oc: objectclass != oc
            and self._is_referencing(oc.fields, objectclass),
            self.objectclasses,
        )

        return list(objectclasses)

    def _is_referencing(self, fields, objectclass):
        fields = filter(self._is_objectclass_field, fields)
        return any(field["objectClassName"] == objectclass.name for field in fields)

    def _get_root(self):
        return list(filter(lambda oc: not self._has_references(oc), self.objectclasses))

    def _has_references(self, objectclass):
        return any(
            self._is_objectclass_field(field)
            and field["objectClassName"] != objectclass.name
            for field in objectclass.fields
        )

    def _is_objectclass_field(self, field):
        return field["type"] == "emsFieldTypeObjectClass"



#
# The MIT License (MIT)
# Copyright (c) 2022 M.O.S.S. Computer Grafik Systeme GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the Software
# is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT,TORT OR OTHERWISE, ARISING FROM, OUT
# OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import json
import logging
import mimetypes
import os
import re
from typing import List, Optional

from requests import Response

from moss.ems.utilities import PaginationQuery

logger = logging.getLogger(__file__)


class EmsLayerException(Exception):
    """Exception for EmsLayer."""


class EmsLayer(object):
    """EmsLayer Class."""

    def __init__(self, name, object_class, project, session, definition=None):
        """Initialize the layer class.

        Args:
            name: the name of the layer
            object_class: an instance of an objectclass
            project: an instance of a project
            session: a session object
            definition: a definition of the layer

        Returns:

        Raises:

        """
        self.name = name
        self.object_class = object_class
        self.project = project

        self._session = session

        if definition is not None:
            self.definition = definition

    def query(
        self, output_format="json", where="1=1", variants=None, geometry=None, **kwargs
    ):
        """Run a paginated query on a layer

        Args:
            output_format: str: the supported output formats of the EMS query Rest call
            where: a SQL-like string
            geometry: am Esri Envelope used to filter the query
            variants: a list of variant IDs to query
            **kwargs: additional a keyworded, variable-length argument list
            please refer to ems documentation

        Returns:
            a pagination query object
        """

        url = "/".join(["rest", self.project, self.object_class, self.name, "query"])

        request_parameters = {"f": output_format, "where": where}

        if variants is not None:
            if isinstance(variants, list):
                variant_ids = ",".join([str(variant_id) for variant_id in variants])
                request_parameters["variants"] = variant_ids
            else:
                request_parameters["variants"] = variants

        #  To ensure compatibility with the old version
        #  In the old version geometry was a boolean to returnGeometry

        if type(geometry) == bool:
            request_parameters["returnGeometry"] = geometry
        else:
            if geometry is not None:
                request_parameters["geometry"] = geometry

        for key, item in kwargs.items():
            request_parameters[key] = item

        return PaginationQuery(self._session, url, self.project, **request_parameters)

    def add_features(self, features, variant_id=None):
        """Add features to a layer

        Args:
            features: a list of features
            variant_id:

        Returns:
            the response dict for the addFeatures request or None on error
        Raises:
            EmsLayerException
        """

        url = "/".join(
            ["rest", self.project, self.object_class, self.name, "addFeatures"]
        )

        if not isinstance(features, list):
            raise EmsLayerException("The features have to be of type 'list'.")

        post_data = {
            "f": "json",
            "features": json.dumps(features),
        }

        # Add the variant parameter if provided
        if variant_id is not None:
            logger.debug("Using variant: variant_id is %s ", variant_id)

            try:
                post_data["variant"] = int(variant_id)
            except TypeError:
                logger.error("The variant_id is not an integer")
                raise EmsLayerException("The variant_id is not an integer")

        logger.debug("Sending request data: %s", post_data)

        response = self._session.post(url, data=post_data)
        response_dict = response.json()

        logger.debug("add_feature: response %s", response_dict)
        error_flag = False

        # addResults could have more than one result sets
        try:
            for i in range(len(response_dict["addResults"])):
                if response_dict["addResults"][i]["success"] is False:
                    error_flag = True
                    error_dictionary = response_dict["addResults"][i]["error"]

                    logger.error("Message: %s", error_dictionary["message"])
                    logger.error("Description: %s", error_dictionary["description"])
                    logger.error("Code: %s", error_dictionary["code"])
        except KeyError:
            error_dictionary = response_dict["error"]
            logger.error("Message: %s", error_dictionary["message"])
            logger.error("Description: %s", error_dictionary["description"])
            logger.error("Code: %s", error_dictionary["code"])
            return error_dictionary

        if error_flag is True:
            logger.info("Not all the features were added successfully.")
        else:
            logger.info("All the features were added successfully.")
        logger.debug("Features:\n%s", json.dumps(features))
        return response_dict

    def update_features(self, features, variant_id=None):
        """Modifies the features of a layer

        Args:
            features: a list of features
            variant_id:

        Returns:
            the response dict for the updateFeatures request or None on error
        Raises:
            EmsLayerException
        """

        if not isinstance(features, list):
            raise EmsLayerException("The features have to be of type 'list'.")

        # firstly get which fields (via objectClass) are editable
        url = "/".join(["rest", self.project, "objectClasses"])
        request_parameters = {"f": "json"}

        response = self._session.get(url, params=request_parameters)
        response_dict = response.json()
        objectclass = next(
            iter(
                filter(
                    lambda x: x["name"] == self.object_class,
                    response_dict["objectClasses"],
                )
            ),
            None,
        )

        # avoid sending fields that are not editable.
        exclude_fields = []

        for field in objectclass["fields"]:
            if field["name"] != "ID":
                if not field["editable"]:
                    exclude_fields.append(field["name"])

        for feature in features:
            for attribute in list(feature["attributes"]):
                if attribute in exclude_fields:
                    feature["attributes"].pop(attribute, None)

        post_data = {"f": "json", "features": json.dumps(features)}
        # Add the variant parameter if provided
        if variant_id is not None:
            logger.debug("Using variant: variant_id is %s ", variant_id)

            try:
                post_data["variant"] = int(variant_id)
            except TypeError:
                logger.error("The variant_id is not an integer")
                raise EmsLayerException("The variant_id is not an integer")

        url = "/".join(
            ["rest", self.project, self.object_class, self.name, "updateFeatures"]
        )

        response = self._session.post(url, data=post_data)
        response_dict = response.json()
        error_flag = False

        # updateResults could have more than one result sets
        try:
            for i in range(len(response_dict["updateResults"])):
                if response_dict["updateResults"][i]["success"] is False:
                    error_flag = True
                    error_dictionary = response_dict["updateResults"][i]["error"]

                    logger.error("Message: %s", error_dictionary["message"])
                    logger.error("Description: %s", error_dictionary["description"])
                    logger.error("Code: %s", error_dictionary["code"])
        except KeyError:
            error_dictionary = response_dict["error"]
            logger.error("Message: %s", error_dictionary["message"])
            logger.error("Description: %s", error_dictionary["description"])
            logger.error("Code: %s", error_dictionary["code"])
            return error_dictionary

        if error_flag is True:
            logger.info("Not all the features were modified successfully.")
        else:
            logger.info("All the features were modified successfully.")
        return response_dict

    def delete_features(self, object_ids, where=None, variant_id=None):
        """Delete the features in the list of objectIds of a layer

        Args:
            object_ids: a list of object IDs to delete the features
            where: a SQL statement

        Returns:
            the response dict for the deleteFeatures request or None on error
        """

        url = "/".join(
            ["rest", self.project, self.object_class, self.name, "deleteFeatures"]
        )

        if isinstance(object_ids, list):
            ids = ",".join([str(object_id) for object_id in object_ids])
        else:
            ids = object_ids

        post_data = {"f": "json", "objectIds": ids}

        if where is not None:
            post_data["where"] = where

        # Add the variant parameter if provided
        if variant_id is not None:
            logger.debug("Using variant: variant_id is %s ", variant_id)

            try:
                post_data["variant"] = int(variant_id)
            except TypeError:
                logger.exception("The variant_id is not an integer")

        response = self._session.post(url, data=post_data)
        response_dict = response.json()
        error_flag = False

        try:
            # deleteResults could have more than one result set
            for i in range(len(response_dict["deleteResults"])):
                if response_dict["deleteResults"][i]["success"] is False:
                    error_flag = True
                    error_dictionary = response_dict["deleteResults"][i]["error"]

                    logger.error("Message: %s", error_dictionary["message"])
                    logger.error("Description: %s", error_dictionary["description"])
                    logger.error("Code: %s", error_dictionary["code"])
        except KeyError:
            error_dictionary = response_dict["error"]
            logger.error("Message: %s", error_dictionary["message"])
            logger.error("Description: %s", error_dictionary["description"])
            logger.error("Code: %s", error_dictionary["code"])
            return error_dictionary

        if error_flag is True:
            logger.info("Not all the features were deleted successfully.")
        else:
            logger.info("All the features were deleted successfully.")
        return response_dict

    def attachments(self, feature_id="1", asUrl=False, appType="master"):
        """
        This will list all the attachment infos for given feature.

        Args:

            feature_id (int): The id of the feature
            asUrl (bool): Return the storage path of the attachment. Using this
                          option is a security risk!
            appType (str): archive/preview/maste(default)
        """

        url = "/".join(
            [
                "rest",
                self.project,
                self.object_class,
                self.name,
                str(feature_id),
                "attachments",
            ]
        )
        get_params = {"f": "json", "asUrl": asUrl, "appType": appType}
        response = self._session.get(url, params=get_params)
        response_dict = response.json()

        logger.debug(response_dict)

        logger.info("Successfully queried the attachments.")
        return response_dict

    def add_attachment(self, feature_id, app_type="master", file_path=""):
        """Adds an attachment to a layer/feature

        not officially supported yet
        """

        url = "/".join(
            [
                "rest",
                self.project,
                self.object_class,
                self.name,
                str(feature_id),
                "addAttachment",
            ]
        )

        # absolute path
        upload_file = os.path.abspath(file_path)
        file_name = os.path.basename(upload_file)

        if not os.path.exists(file_path):
            raise EmsLayerException("The path {} does not exist".format(upload_file))

        files = {"attachment": (file_name, open(upload_file, "rb"))}

        post_data = {"f": "json", "apptype": app_type}
        response_dict = self._session.post(url, files=files, data=post_data).json()

        # results could have more than one result set
        try:
            if response_dict["addAttachmentResult"]["success"] is False:
                error_dictionary = response_dict["addAttachmentResult"]["error"]

                logger.error("Message: %s", error_dictionary["message"])
                logger.error("Description: %s", error_dictionary["description"])
                logger.error("Code: %s", error_dictionary["code"])
                return None
        except KeyError:
            error_dictionary = response_dict["error"]
            logger.error("Message: %s", error_dictionary["message"])
            logger.error("Description: %s", error_dictionary["description"])
            logger.error("Code: %s", error_dictionary["code"])
            return None

        logger.info("The attachment was added successfully.")
        return response_dict

    def update_attachment(self, feature_id, attachment_id, file_path):
        #  type: (int, int, str) -> bool
        """
        Update an attachment with a new one that will be uploaded

        Args:
            feature_id (int): The ID of the feature with attachment
            attachment_id (int): The ID of the attachment to be updated
            file_path (str): The Path to the new attachment file

        Returns:
            status (bool): The Status of the operation
        """

        url = "/".join(
            [
                "rest",
                self.project,
                self.object_class,
                self.name,
                str(feature_id),
                "updateAttachment",
            ]
        )

        # absolute path
        upload_file = os.path.abspath(file_path)
        file_name = os.path.basename(upload_file)

        if not os.path.exists(file_path):
            raise EmsLayerException("The path {} does not exist".format(upload_file))

        files = {"attachment": (file_name, open(upload_file, "rb"))}

        post_data = {"f": "json", "attachmentId": attachment_id}
        response_dict = self._session.post(url, files=files, data=post_data).json()

        # results could have more than one result set
        try:
            if response_dict["updateAttachmentResult"]["success"] is False:
                error_dictionary = response_dict["updateAttachmentResult"]["error"]

                logger.error("Message: %s", error_dictionary["message"])
                logger.error("Description: %s", error_dictionary["description"])
                logger.error("Code: %s", error_dictionary["code"])
                return False
        except KeyError:
            error_dictionary = response_dict["error"]
            logger.error("Message: %s", error_dictionary["message"])
            logger.error("Description: %s", error_dictionary["description"])
            logger.error("Code: %s", error_dictionary["code"])
            return False

        logger.info("The attachment was updated successfully.")
        return True

    def register_attachment(
        self,
        feature_id,
        app_type="master",
        file_path="",
        fixed_path=True,
        content_type=None,
    ):
        """
        Registers an attachment without uploading a file.
        This function is used when attachments are already in the right place.
        This can e.g. be the case with a migration. Please note that the entry
        must be an absolute path

        Args:
            feature_id (int): The ID of the feature owning the attachment
            app_type (str): The apptype archive or master
            file_path (str): The path of the file to register
            fixed_path (bool): If True the base_path path is not prefixed.
            content_type (str): The Content type of the file.

        Returns:
            bool: The result of the register attachment
        """
        logger.info("Registering attachment.")

        url = "/".join(
            [
                "rest",
                self.project,
                self.object_class,
                self.name,
                feature_id,
                "registerAttachment",
            ]
        )
        register_file = os.path.abspath(file_path)
        logger.debug("File path is {0}".format(register_file))

        if not os.path.exists(file_path):
            raise EmsLayerException("The path {} does not exist".format(register_file))

        file_size = os.path.getsize(register_file)
        logger.debug("File size is {0}".format(file_size))

        # Guess mimetype if not defined
        if app_type is None:
            app_type = mimetypes.guess_type(register_file)[0]
        logger.debug("Detected mimetype is {0}".format(app_type))
        # If the mimetype is still none, raise an error
        if app_type is None:
            raise EmsLayerException("Can not detect mimetype, please specify it.")

        post_data = {
            "f": "json",
            "apptype": app_type,
            "filePath": file_path,
            "fixedPath": fixed_path,
            "contentType": content_type,
            "size": file_size,
        }
        logger.debug("Post data {0}".format(post_data))
        response_dict = self._session.post(url, data=post_data).json()

        # results could have more than one result set
        try:
            if response_dict["addAttachmentResult"]["success"] is False:
                error_dictionary = response_dict["addAttachmentResult"]["error"]

                logger.error("Message: %s", error_dictionary["message"])
                logger.error("Description: %s", error_dictionary["description"])
                logger.error("Code: %s", error_dictionary["code"])
                return False
        except KeyError:
            error_dictionary = response_dict["error"]
            logger.error("Message: %s", error_dictionary["message"])
            logger.error("Description: %s", error_dictionary["description"])
            logger.error("Code: %s", error_dictionary["code"])
            return False

        logger.info("The attachment was registered successfully.")
        return True

    def delete_attachment(self, feature_id, attachment_ids):
        #  type: (int, List[int]) -> bool
        """
        Enables the deletion of several existing files or ZIP archives of a feature.

        Args:
            feature_id (int): The ID of the feature
            attachments_id (list): The list of attachments id to delete

        Return:
            result (bool): If the operation war succesfull
        """

        if type(attachment_ids) is list:
            attachments_list = ",".join(
                [str(attachment_id) for attachment_id in attachment_ids]
            )
            logger.debug("Deleting {0}".format(attachments_list))
        else:
            raise EmsLayerException("The attachments id should be a list of int")

        url = "/".join(
            [
                "rest",
                self.project,
                self.object_class,
                self.name,
                str(feature_id),
                "deleteAttachment",
            ]
        )

        post_data = {"f": "json", "attachmentIds": attachments_list}
        response_dict = self._session.post(url, data=post_data).json()

        # results could have more than one result set
        try:
            for attachment in response_dict["deleteAttachmentResult"]:
                if attachment["success"] is False:
                    error_dictionary = response_dict["deleteAttachmentResult"]["error"]

                    logger.error("Message: %s", error_dictionary["message"])
                    logger.error("Description: %s", error_dictionary["description"])
                    logger.error("Code: %s", error_dictionary["code"])
                    return False
                else:
                    logger.info(
                        "The attachment '%s' was deleted successfully.",
                        attachment["objectId"],
                    )
        except KeyError:
            error_dictionary = response_dict["error"]
            logger.error("Message: %s", error_dictionary["message"])
            logger.error("Description: %s", error_dictionary["description"])
            logger.error("Code: %s", error_dictionary["code"])
            return False

        if len(response_dict["deleteAttachmentResult"]) == 1:
            logger.info("The attachment was deleted successfully.")
        else:
            logger.info("The attachments were deleted successfully.")
        return True

    def download_attachment(self, feature_id, attachment_id, output_path=None):
        # type: (int, int, Optional[str]) -> Optional[bool]
        """
        Download the specified attachment

        Args:
            feature_id (int): The ID of the feature owning the attachment
            attachment_id (int): The attachments ID
            output_path (str): Output directory where the attachment will be saved
        Returns:
            bool: The result of the download attachment
        """

        logger.info(
            "Downloading attachment {0} for feature {1}".format(
                attachment_id, feature_id
            )
        )

        url = "/".join(
            [
                "rest",
                self.project,
                self.object_class,
                self.name,
                str(feature_id),
                "attachments",
                str(attachment_id),
            ]
        )

        response = self._session.get(url)  # type: Response
        content_type = response.headers.get("Content-Type")  # type: Optional[str]
        if content_type is not None:
            if "application/json" not in content_type:
                logger.info("Headers {0}".format(response.headers))

                content_disposition = response.headers.get("Content-Disposition")
                if content_disposition:
                    filename = re.findall("filename=(.+)", content_disposition)[0]
                else:
                    filename = "no_name"
                logger.info("Saving attachment to {0}".format(filename))

                if output_path:
                    if not os.path.exists(output_path):
                        logger.error("The path {0} does not exist".format(output_path))
                        return False

                    attachment_path = os.path.join(output_path, filename)

                else:
                    attachment_path = filename

                logger.debug("Saving attachment to {0}".format(attachment_path))
                with open(attachment_path, "wb") as output_file:
                    output_file.write(response.content)
                return True
            else:
                logger.error("The attachment url {0} does not exist".format(url))
                return False

    def __repr__(self):
        return self.name



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020  by M.O.S.S. Computer Grafik Systeme GmbH
#                         Hohenbrunner Weg 13
#                         D-82024 Taufkirchen
#                         http://www.moss.de#

# -*- coding: utf-8 -*-

import sys
from getpass import getpass

this = sys.modules[__name__]

this.default_max_try = 5
this.default_interactive = True
this.default_clear_before_return = False
this.default_is_secure = False
this.default_required = True


def set_prompt_defaults(
    max_try=this.default_max_try,
    required=this.default_required,
    is_secure=this.default_is_secure,
    interactive=this.default_interactive,
    clear_before_return=this.default_clear_before_return,
):
    this.default_max_try = max_try
    this.default_required = required
    this.default_is_secure = is_secure
    this.default_interactive = interactive
    this.default_clear_before_return = clear_before_return


def prompt(
    message,
    default=None,
    max_try=None,
    required=None,
    is_secure=None,
    is_bool=False,
    interactive=None,
    clear_before_return=None,
):
    if max_try is None:
        max_try = this.default_max_try

    if required is None:
        required = this.default_required

    if is_secure is None:
        is_secure = this.default_is_secure

    if interactive is None:
        interactive = this.default_interactive

    if clear_before_return is None:
        clear_before_return = this.default_clear_before_return

    result = None

    if interactive is False:
        if required:
            exit_when_empty(
                result,
                "Expected Parameter {parameter} to be defined:".format(
                    parameter=message
                ),
            )

        return default

    if default is not None:
        message = "{message} ({default})".format(message=message, default=default)

    message = message + ": "

    while max_try > 0 and result is None:
        max_try = max_try - 1

        if is_secure:
            result = getpass(message)
        else:
            result = input(message)

            if result == "" and default is not None:
                result = default
            elif is_bool:
                result = result.lower() in ["y", "true", "yes"]

        result = result.strip()

        if required and result == "":
            result = None

    if clear_before_return:
        print("\033[A                                   \033[A")

    return result


def exit_when_empty(value, message):
    if value is None or value == "":
        print(message)
        sys.exit(1)
